#!/usr/bin/env python
# @author Martin Karkowski
# @email m.karkowski@zema.de

from .baseModule import BaseModule
from .genericModule import NopeGenericModule

__all__ = ["BaseModule", "NopeGenericModule"]
